from ._linear import LinearRegression
from ._lasso import LassoRegression
from ._dtree import DecisionTreeRegression
from ._rforest import RandomForestRegression
from ._nn import NeuralNetsRegression
from ._nn_wrapper import NeuralNetsWrapper
from ._kneighbors import KNeighborsRegression
from ._svr import SVRegression



import os
import random
import numpy as np
import tensorflow as tf
import warnings

# Class for global settings
class Settings:
    def __init__(self, update: dict = None):
        # Defaults
        self._verbosity = 0
        self._random_state = None
        self._test_size = 0.3
        self._num_configs_saved = 5

        # If a dictionary of key/value pairs is given,
        # update settings
        if update != None:
            for key, value in update.items():
                setattr(self, key, value)

        if self._verbosity <= 1:
            os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
            tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)

            warnings.simplefilter(action="ignore", category=Warning)
            warnings.simplefilter(action="ignore", category=FutureWarning)

        if self._random_state != None:
            os.environ["PYTHONHASHSEED"] = str(self._random_state)
            random.seed(self._random_state)
            np.random.seed(self._random_state)
            tf.compat.v1.set_random_seed(self._random_state)

            # Deterministic tensorflow
            os.environ["TF_DETERMINISTIC_OPS"] = "1"
            os.environ["TF_CUBNN_DETERMINISTIC"] = "1"

            session_conf = tf.compat.v1.ConfigProto(
                intra_op_parallelism_threads=1, inter_op_parallelism_threads=1
            )
            sess = tf.compat.v1.Session(
                graph=tf.compat.v1.get_default_graph(), config=session_conf
            )
            tf.compat.v1.keras.backend.set_session(sess)

    # Getters
    @property
    def verbosity(self) -> int:
        return self._verbosity

    @property
    def random_state(self) -> int:
        return self._random_state

    @property
    def test_size(self) -> float:
        return self._test_size

    @property
    def num_configs_saved(self) -> int:
        return self._num_configs_saved

    # Setters
    @verbosity.setter
    def verbosity(self, verbosity: int):
        assert isinstance(verbosity, int)
        assert verbosity >= 0
        self._verbosity = verbosity

    @random_state.setter
    def random_state(self, random_state: int):
        assert random_state == None or random_state >= 0
        self._random_state = random_state

    @test_size.setter
    def test_size(self, test_size: float):
        assert isinstance(test_size, float)
        assert test_size >= 0.0 and test_size < 1.0
        self._test_size = test_size

    @num_configs_saved.setter
    def num_configs_saved(self, num_configs_saved: int):
        assert num_configs_saved > 0
        self._num_configs_saved = num_configs_saved


# Initialization function for global settings
def init(settings_changes: dict = None):
    global values
    values = Settings(settings_changes)
    return values



from pyMAISE.methods._nn_wrapper import NeuralNetsWrapper


class NeuralNetsRegression:
    def __init__(self, parameters: dict = None):
        # Model parameters
        # Sequential
        self._num_layers = None
        self._name = None
        self._dropout = False
        self._rate = 0.5
        self._loss = None
        self._metrics = None
        self._loss_weights = None
        self._weighted_metrics = None
        self._run_eagerly = False
        self._steps_per_execution = None
        self._jit_compile = True  # Used for both compile and adam
        self._batch_size = None
        self._validation_batch_size = None
        self._shuffle = True
        self._callbacks = None
        self._validation_split = 0.0
        self._epochs = 1
        self._warm_start = False

        # Starting Layer
        self._start_num_nodes = None
        self._start_activation = None
        self._start_use_bias = True
        self._start_kernel_initializer = "glorot_uniform"
        self._start_bias_initializer = "zeros"
        self._start_kernel_regularizer = None
        self._start_bias_regularizer = None
        self._start_activity_regularizer = None
        self._start_kernel_constraint = None
        self._start_bias_constraint = None
        self._input_dim = None

        # Middle Layers
        self._mid_num_node_strategy = None
        self._mid_activation = None
        self._mid_use_bias = True
        self._mid_kernel_initializer = "glorot_uniform"
        self._mid_bias_initializer = "zeros"
        self._mid_kernel_regularizer = None
        self._mid_bias_regularizer = None
        self._mid_activity_regularizer = None
        self._mid_kernel_constraint = None
        self._mid_bias_constraint = None

        # Ending Layer
        self._end_num_nodes = None
        self._end_activation = None
        self._end_use_bias = True
        self._end_kernel_initializer = "glorot_uniform"
        self._end_bias_initializer = "zeros"
        self._end_kernel_regularizer = None
        self._end_bias_regularizer = None
        self._end_activity_regularizer = None
        self._end_kernel_constraint = None
        self._end_bias_constraint = None

        # Optimizer
        self._optimizer = "adam"

        # Adam
        self._learning_rate = 0.001
        self._beta_1 = 0.9
        self._beta_2 = 0.999
        self._epsilon = 1e-7
        self._amsgrad = False
        self._clipnorm = None
        self._clipvalue = None
        self._global_clipnorm = None

        # Change if user provided changes in dictionary
        if parameters != None:
            for key, value in parameters.items():
                setattr(self, key, value)

    # ===========================================================
    # Methods
    def regressor(self):
        return NeuralNetsWrapper(
            num_layers=self._num_layers,
            name=self._name,
            dropout=self._dropout,
            rate=self._rate,
            loss=self._loss,
            metrics=self._metrics,
            loss_weights=self._loss_weights,
            weighted_metrics=self._weighted_metrics,
            run_eagerly=self._run_eagerly,
            steps_per_execution=self._steps_per_execution,
            batch_size=self._batch_size,
            validation_batch_size=self._validation_batch_size,
            shuffle=self._shuffle,
            callbacks=self._callbacks,
            validation_split=self._validation_split,
            epochs=self._epochs,
            warm_start=self._warm_start,
            start_num_nodes=self._start_num_nodes,
            start_activation=self._start_activation,
            start_use_bias=self._start_use_bias,
            start_kernel_initializer=self._start_kernel_initializer,
            start_bias_initializer=self._start_bias_initializer,
            start_kernel_regularizer=self._start_kernel_regularizer,
            start_bias_regularizer=self._start_bias_regularizer,
            start_activity_regularizer=self._start_activity_regularizer,
            start_kernel_constraint=self._start_kernel_constraint,
            start_bias_constraint=self._start_bias_constraint,
            input_dim=self._input_dim,
            mid_num_node_strategy=self._mid_num_node_strategy,
            mid_activation=self._mid_activation,
            mid_use_bias=self._mid_use_bias,
            mid_kernel_initializer=self._mid_kernel_initializer,
            mid_bias_initializer=self._mid_bias_initializer,
            mid_kernel_regularizer=self._mid_kernel_regularizer,
            mid_bias_regularizer=self._mid_bias_regularizer,
            mid_activity_regularizer=self._mid_activity_regularizer,
            mid_kernel_constraint=self._mid_kernel_constraint,
            mid_bias_constraint=self._mid_bias_constraint,
            end_num_nodes=self._end_num_nodes,
            end_activation=self._end_activation,
            end_use_bias=self._end_use_bias,
            end_kernel_initializer=self._end_kernel_initializer,
            end_bias_initializer=self._end_bias_initializer,
            end_kernel_regularizer=self._end_kernel_regularizer,
            end_bias_regularizer=self._end_bias_regularizer,
            end_activity_regularizer=self._end_activity_regularizer,
            end_kernel_constraint=self._end_kernel_constraint,
            end_bias_constraint=self._end_bias_constraint,
            optimizer=self._optimizer,
            learning_rate=self._learning_rate,
            beta_1=self._beta_1,
            beta_2=self._beta_2,
            epsilon=self._epsilon,
            amsgrad=self._amsgrad,
            clipnorm=self._clipnorm,
            clipvalue=self._clipvalue,
            global_clipnorm=self._global_clipnorm,
            jit_compile=self._jit_compile,
        )

    # ===========================================================
    # Getters
    @property
    def num_layers(self):
        return self._num_layers

    @property
    def name(self):
        return self._name

    @property
    def dropout(self):
        return self._dropout

    @property
    def rate(self):
        return self._rate

    @property
    def loss(self):
        return self._loss

    @property
    def metrics(self):
        return self._metrics

    @property
    def loss_weights(self):
        return self._loss_weights

    @property
    def run_eagerly(self):
        return self._run_eagerly

    @property
    def steps_per_execution(self):
        return self._steps_per_execution

    @property
    def jit_compile(self):
        return self._jit_compile

    @property
    def batch_size(self):
        return self._batch_size

    @property
    def validation_batch_size(self):
        return self._validation_batch_size

    @property
    def shuffle(self):
        return self._shuffle

    @property
    def callbacks(self):
        return self._callbacks

    @property
    def validation_split(self):
        return self._validation_split

    @property
    def epochs(self):
        return self._epochs

    @property
    def warm_start(self):
        return self._warm_start

    @property
    def start_num_nodes(self):
        return self._start_num_nodes

    @property
    def start_activation(self):
        return self._start_activation

    @property
    def start_use_bias(self):
        return self._start_use_bias

    @property
    def start_kernel_initializer(self):
        return self._start_kernel_initializer

    @property
    def start_bias_initializer(self):
        return self._start_bias_initializer

    @property
    def start_kernel_regularizer(self):
        return self._start_kernel_regularizer

    @property
    def start_bias_regularizer(self):
        return self._start_bias_regularizer

    @property
    def start_activity_regularizer(self):
        return self._start_activity_regularizer

    @property
    def start_kernel_constraint(self):
        return self._start_kernel_constraint

    @property
    def start_bias_constraint(self):
        return self._start_bias_constraint

    @property
    def input_dim(self):
        return self._input_dim

    @property
    def mid_num_node_strategy(self):
        return self._mid_num_node_strategy

    @property
    def mid_activation(self):
        return self._mid_activation

    @property
    def mid_use_bias(self):
        return self._mid_use_bias

    @property
    def mid_kernel_initializer(self):
        return self._mid_kernel_initializer

    @property
    def mid_bias_initializer(self):
        return self._mid_bias_initializer

    @property
    def mid_kernel_regularizer(self):
        return self._mid_kernel_regularizer

    @property
    def mid_bias_regularizer(self):
        return self._mid_bias_regularizer

    @property
    def mid_activity_regularizer(self):
        return self._mid_activity_regularizer

    @property
    def mid_kernel_constraint(self):
        return self._mid_kernel_constraint

    @property
    def mid_bias_constraint(self):
        return self._mid_bias_constraint

    @property
    def end_num_nodes(self):
        return self._end_num_nodes

    @property
    def end_activation(self):
        return self._end_activation

    @property
    def end_use_bias(self):
        return self._end_use_bias

    @property
    def end_kernel_initializer(self):
        return self._end_kernel_initializer

    @property
    def end_bias_initializer(self):
        return self._end_bias_initializer

    @property
    def end_kernel_regularizer(self):
        return self._end_kernel_regularizer

    @property
    def end_bias_regularizer(self):
        return self._end_bias_regularizer

    @property
    def end_activity_regularizer(self):
        return self._end_activity_regularizer

    @property
    def end_kernel_constraint(self):
        return self._start_end_constraint

    @property
    def end_bias_constraint(self):
        return self._end_bias_constraint

    @property
    def optimizer(self):
        return self._optimizer

    @property
    def learning_rate(self):
        return self._learning_rate

    @property
    def beta_1(self):
        return self._beta_1

    @property
    def beta_2(self):
        return self._beta_2

    @property
    def epsilon(self):
        return self._epsilon

    @property
    def amsgrad(self):
        return self._amsgrad

    @property
    def clipnorm(self):
        return self._clipnorm

    @property
    def clipvalue(self):
        return self._clipvalue

    @property
    def global_clipnorm(self):
        return self._global_clipnorm

    # ===========================================================
    # Setters
    @num_layers.setter
    def num_layers(self, num_layers):
        self._num_layers = num_layers

    @name.setter
    def name(self, name):
        self._name = name

    @dropout.setter
    def dropout(self, dropout):
        self._dropout = dropout

    @rate.setter
    def rate(self, rate):
        self._rate = rate

    @loss.setter
    def loss(self, loss):
        self._loss = loss

    @metrics.setter
    def metrics(self, metrics):
        self._metrics = metrics

    @loss_weights.setter
    def loss_weights(self, loss_weights):
        self._loss_weights = loss_weights

    @run_eagerly.setter
    def run_eagerly(self, run_eagerly):
        self._run_eagerly = run_eagerly

    @steps_per_execution.setter
    def steps_per_execution(self, steps_per_execution):
        self._steps_per_execution = steps_per_execution

    @jit_compile.setter
    def jit_compile(self, jit_compile):
        self._jit_compile = jit_compile

    @batch_size.setter
    def batch_size(self, batch_size):
        self._batch_size = batch_size

    @validation_batch_size.setter
    def validation_batch_size(self, validation_batch_size):
        self._validation_batch_size = validation_batch_size

    @shuffle.setter
    def shuffle(self, shuffle):
        self._shuffle = shuffle

    @callbacks.setter
    def callbacks(self, callbacks):
        self._callbacks = callbacks

    @validation_split.setter
    def validation_split(self, validation_split):
        self._validation_split = validation_split

    @epochs.setter
    def epochs(self, epochs):
        self._epochs = epochs

    @warm_start.setter
    def warm_start(self, warm_start):
        self._warm_start = warm_start

    @start_num_nodes.setter
    def start_num_nodes(self, start_num_nodes):
        self._start_num_nodes = start_num_nodes

    @start_activation.setter
    def start_activation(self, start_activation):
        self._start_activation = start_activation

    @start_use_bias.setter
    def start_use_bias(self, start_use_bias):
        self._start_use_bias = start_use_bias

    @start_kernel_initializer.setter
    def start_kernel_initializer(self, start_kernel_initializer):
        self._start_kernel_initializer = start_kernel_initializer

    @start_bias_initializer.setter
    def start_bias_initializer(self, start_bias_initializer):
        self._start_bias_initializer = start_bias_initializer

    @start_kernel_regularizer.setter
    def start_kernel_regularizer(self, start_kernel_regularizer):
        self._start_kernel_regularizer = start_kernel_regularizer

    @start_bias_regularizer.setter
    def start_bias_regularizer(self, start_bias_regularizer):
        self._start_bias_regularizer = start_bias_regularizer

    @start_activity_regularizer.setter
    def start_activity_regularizer(self, start_activity_regularizer):
        self._start_activity_regularizer = start_activity_regularizer

    @start_kernel_constraint.setter
    def start_kernel_constraint(self, start_kernel_constraint):
        self._start_kernel_constraint = start_kernel_constraint

    @start_bias_constraint.setter
    def start_bias_constraint(self, start_bias_constraint):
        self._start_bias_constraint = start_bias_constraint

    @input_dim.setter
    def input_dim(self, input_dim):
        self._input_dim = input_dim

    @mid_num_node_strategy.setter
    def mid_num_node_strategy(self, mid_num_node_strategy):
        assert isinstance(mid_num_node_strategy, str) or isinstance(
            mid_num_node_strategy, callable
        )
        self._mid_num_node_strategy = mid_num_node_strategy

    @mid_activation.setter
    def mid_activation(self, mid_activation):
        self._mid_activation = mid_activation

    @mid_use_bias.setter
    def mid_use_bias(self, mid_use_bias):
        self._mid_use_bias = mid_use_bias

    @mid_kernel_initializer.setter
    def mid_kernel_initializer(self, mid_kernel_initializer):
        self._mid_kernel_initializer = mid_kernel_initializer

    @mid_bias_initializer.setter
    def mid_bias_initializer(self, mid_bias_initializer):
        self._mid_bias_initializer = mid_bias_initializer

    @mid_kernel_regularizer.setter
    def mid_kernel_regularizer(self, mid_kernel_regularizer):
        self._mid_kernel_regularizer = mid_kernel_regularizer

    @mid_bias_regularizer.setter
    def mid_bias_regularizer(self, mid_bias_regularizer):
        self._mid_bias_regularizer = mid_bias_regularizer

    @mid_activity_regularizer.setter
    def mid_activity_regularizer(self, mid_activity_regularizer):
        self._mid_activity_regularizer = mid_activity_regularizer

    @mid_kernel_constraint.setter
    def mid_kernel_constraint(self, mid_kernel_constraint):
        self._mid_kernel_constraint = mid_kernel_constraint

    @mid_bias_constraint.setter
    def mid_bias_constraint(self, mid_bias_constraint):
        self._mid_bias_constraint = mid_bias_constraint

    @end_num_nodes.setter
    def end_num_nodes(self, end_num_nodes):
        self._end_num_nodes = end_num_nodes

    @end_activation.setter
    def end_activation(self, end_activation):
        self._end_activation = end_activation

    @end_use_bias.setter
    def end_use_bias(self, end_use_bias):
        self._end_use_bias = end_use_bias

    @end_kernel_initializer.setter
    def end_kernel_initializer(self, end_kernel_initializer):
        self._end_kernel_initializer = end_kernel_initializer

    @end_bias_initializer.setter
    def end_bias_initializer(self, end_bias_initializer):
        self._end_bias_initializer = end_bias_initializer

    @end_kernel_regularizer.setter
    def end_kernel_regularizer(self, end_kernel_regularizer):
        self._end_kernel_regularizer = end_kernel_regularizer

    @end_bias_regularizer.setter
    def end_bias_regularizer(self, end_bias_regularizer):
        self._end_bias_regularizer = end_bias_regularizer

    @end_activity_regularizer.setter
    def end_activity_regularizer(self, end_activity_regularizer):
        self._end_activity_regularizer = end_activity_regularizer

    @end_kernel_constraint.setter
    def end_kernel_constraint(self, end_kernel_constraint):
        self._start_end_constraint = end_kernel_constraint

    @end_bias_constraint.setter
    def end_bias_constraint(self, end_bias_constraint):
        self._end_bias_constraint = end_bias_constraint

    @optimizer.setter
    def optimizer(self, optimizer):
        self._optimizer = optimizer

    @learning_rate.setter
    def learning_rate(self, learning_rate):
        self._learning_rate = learning_rate

    @beta_1.setter
    def beta_1(self, beta_1):
        self._beta_1 = beta_1

    @beta_2.setter
    def beta_2(self, beta_2):
        self._beta_2 = beta_2

    @epsilon.setter
    def epsilon(self, epsilon):
        self._epsilon = epsilon

    @amsgrad.setter
    def amsgrad(self, amsgrad):
        self._amsgrad = amsgrad

    @clipnorm.setter
    def clipnorm(self, clipnorm):
        self._clipnorm = clipnorm

    @clipvalue.setter
    def clipvalue(self, clipvalue):
        self._clipvalue = clipvalue

    @global_clipnorm.setter
    def global_clipnorm(self, global_clipnorm):
        self._global_clipnorm = global_clipnorm



from pyMAISE.preprocessing import PreProcesser
from pkg_resources import resource_filename

# Get full pyMAISE data file path
def get_full_path(path: str):
    return resource_filename("pyMAISE", path)


# Load benchmark BWR cross section data
def load_xs():
    return PreProcesser(get_full_path("data/xs.csv"), slice(0, -1), slice(-1, None))


# Load benchmark MIT reactor data
def load_MITR():
    return PreProcesser(
        [get_full_path("data/crx.csv"), get_full_path("data/powery.csv")],
    )


# Load benchmark fuel perfromance data
def load_fp():
    return PreProcesser(
        [get_full_path("data/fp_inp.csv"), get_full_path("data/fp_out.csv")]
    )


# Load benchmark fuel centerline temperature data
def load_heat():
    return PreProcesser(get_full_path("data/heat.csv"), slice(0, -1), slice(-1, None))



import pyMAISE.settings as settings

from math import ceil
from keras.models import Sequential
from keras.layers import Dense, Dropout
from tensorflow.keras.optimizers import Adam
from scikeras.wrappers import BaseEstimator, KerasRegressor


class NeuralNetsWrapper(BaseEstimator):
    def __init__(
        self,
        num_layers,
        name,
        dropout,
        rate,
        loss,
        metrics,
        loss_weights,
        weighted_metrics,
        run_eagerly,
        steps_per_execution,
        batch_size,
        validation_batch_size,
        shuffle,
        callbacks,
        validation_split,
        epochs,
        warm_start,
        start_num_nodes,
        start_activation,
        start_use_bias,
        start_kernel_initializer,
        start_bias_initializer,
        start_kernel_regularizer,
        start_bias_regularizer,
        start_activity_regularizer,
        start_kernel_constraint,
        start_bias_constraint,
        input_dim,
        mid_num_node_strategy,
        mid_activation,
        mid_use_bias,
        mid_kernel_initializer,
        mid_bias_initializer,
        mid_kernel_regularizer,
        mid_bias_regularizer,
        mid_activity_regularizer,
        mid_kernel_constraint,
        mid_bias_constraint,
        end_num_nodes,
        end_activation,
        end_use_bias,
        end_kernel_initializer,
        end_bias_initializer,
        end_kernel_regularizer,
        end_bias_regularizer,
        end_activity_regularizer,
        end_kernel_constraint,
        end_bias_constraint,
        optimizer,
        learning_rate,
        beta_1,
        beta_2,
        epsilon,
        amsgrad,
        clipnorm,
        clipvalue,
        global_clipnorm,
        jit_compile,
    ):
        # Sequential
        self._num_layers = num_layers
        self._name = name
        self._dropout = dropout
        self._rate = rate
        self._loss = loss
        self._metrics = metrics
        self._loss_weights = loss_weights
        self._weighted_metrics = weighted_metrics
        self._run_eagerly = run_eagerly
        self._steps_per_execution = steps_per_execution
        self._jit_compile = jit_compile  # Used for both compile and adam
        self._batch_size = batch_size
        self._validation_batch_size = validation_batch_size
        self._shuffle = shuffle
        self._callbacks = callbacks
        self._validation_split = validation_split
        self._epochs = epochs
        self._warm_start = warm_start

        # Starting Layer
        self._start_num_nodes = start_num_nodes
        self._start_activation = start_activation
        self._start_use_bias = start_use_bias
        self._start_kernel_initializer = start_kernel_initializer
        self._start_bias_initializer = start_bias_initializer
        self._start_kernel_regularizer = start_kernel_regularizer
        self._start_bias_regularizer = start_bias_regularizer
        self._start_activity_regularizer = start_activity_regularizer
        self._start_kernel_constraint = start_kernel_constraint
        self._start_bias_constraint = start_bias_constraint
        self._input_dim = input_dim

        # Middle Layers
        self._mid_num_node_strategy = mid_num_node_strategy
        self._mid_activation = mid_activation
        self._mid_use_bias = mid_use_bias
        self._mid_kernel_initializer = mid_kernel_initializer
        self._mid_bias_initializer = mid_bias_initializer
        self._mid_kernel_regularizer = mid_kernel_regularizer
        self._mid_bias_regularizer = mid_bias_regularizer
        self._mid_activity_regularizer = mid_activity_regularizer
        self._mid_kernel_constraint = mid_kernel_constraint
        self._mid_bias_constraint = mid_bias_constraint

        # Ending Layer
        self._end_num_nodes = end_num_nodes
        self._end_activation = end_activation
        self._end_use_bias = end_use_bias
        self._end_kernel_initializer = end_kernel_initializer
        self._end_bias_initializer = end_bias_initializer
        self._end_kernel_regularizer = end_kernel_regularizer
        self._end_bias_regularizer = end_bias_regularizer
        self._end_activity_regularizer = end_activity_regularizer
        self._end_kernel_constraint = end_kernel_constraint
        self._end_bias_constraint = end_bias_constraint

        # Optimizer
        self._optimizer = optimizer

        # Adam
        self._learning_rate = learning_rate
        self._beta_1 = beta_1
        self._beta_2 = beta_2
        self._epsilon = epsilon
        self._amsgrad = amsgrad
        self._clipnorm = clipnorm
        self._clipvalue = clipvalue
        self._global_clipnorm = global_clipnorm

        self = self.NeuralNetsRegressor()

    def constant(self, num_layers: int, start_num_nodes: int, end_num_nodes: int):
        layers = [start_num_nodes] * num_layers
        layers[-1] = end_num_nodes
        return layers

    def linear(self, num_layers: int, start_num_nodes: int, end_num_nodes: int):
        layers = []

        # Determine slope
        increment = (end_num_nodes - start_num_nodes) / (num_layers - 1)

        # Starting number of nodes
        nodes = start_num_nodes

        # Fill layers with starting nodes and increment by slope
        for i in range(1, num_layers + 1):
            layers.append(ceil(nodes))
            nodes = nodes + increment

        return layers

    def fit(self, X, y):
        return self._model.fit(X, y)

    def predict(self, X):
        return self._model.predict(X)

    def score(self, X, y_true):
        return self._model.score(X, y_true)

    def get_params(self, deep=True):
        return {
            "num_layers": self._num_layers,
            "name": self._name,
            "dropout": self._dropout,
            "rate": self._rate,
            "loss": self._loss,
            "metrics": self._metrics,
            "loss_weights": self._loss_weights,
            "weighted_metrics": self._weighted_metrics,
            "run_eagerly": self._run_eagerly,
            "steps_per_execution": self._steps_per_execution,
            "batch_size": self._batch_size,
            "validation_batch_size": self._validation_batch_size,
            "shuffle": self._shuffle,
            "callbacks": self._callbacks,
            "validation_split": self._validation_split,
            "epochs": self._epochs,
            "warm_start": self._warm_start,
            "start_num_nodes": self._start_num_nodes,
            "start_activation": self._start_activation,
            "start_use_bias": self._start_use_bias,
            "start_kernel_initializer": self._start_kernel_initializer,
            "start_bias_initializer": self._start_bias_initializer,
            "start_kernel_regularizer": self._start_kernel_regularizer,
            "start_bias_regularizer": self._start_bias_regularizer,
            "start_activity_regularizer": self._start_activity_regularizer,
            "start_kernel_constraint": self._start_kernel_constraint,
            "start_bias_constraint": self._start_bias_constraint,
            "input_dim": self._input_dim,
            "mid_num_node_strategy": self._mid_num_node_strategy,
            "mid_activation": self._mid_activation,
            "mid_use_bias": self._mid_use_bias,
            "mid_kernel_initializer": self._mid_kernel_initializer,
            "mid_bias_initializer": self._mid_bias_initializer,
            "mid_kernel_regularizer": self._mid_kernel_regularizer,
            "mid_bias_regularizer": self._mid_bias_regularizer,
            "mid_activity_regularizer": self._mid_activity_regularizer,
            "mid_kernel_constraint": self._mid_kernel_constraint,
            "mid_bias_constraint": self._mid_bias_constraint,
            "end_num_nodes": self._end_num_nodes,
            "end_activation": self._end_activation,
            "end_use_bias": self._end_use_bias,
            "end_kernel_initializer": self._end_kernel_initializer,
            "end_bias_initializer": self._end_bias_initializer,
            "end_kernel_regularizer": self._end_kernel_regularizer,
            "end_bias_regularizer": self._end_bias_regularizer,
            "end_activity_regularizer": self._end_activity_regularizer,
            "end_kernel_constraint": self._end_kernel_constraint,
            "end_bias_constraint": self._end_bias_constraint,
            "optimizer": self._optimizer,
            "learning_rate": self._learning_rate,
            "beta_1": self._beta_1,
            "beta_2": self._beta_2,
            "epsilon": self._epsilon,
            "amsgrad": self._amsgrad,
            "clipnorm": self._clipnorm,
            "clipvalue": self._clipvalue,
            "global_clipnorm": self._global_clipnorm,
            "jit_compile": self._jit_compile,
        }

    def set_params(self, **parameters):
        # Change if user provided changes in dictionary
        if parameters != None:
            for key, value in parameters.items():
                setattr(self, key, value)

        self = self.NeuralNetsRegressor()

        return self

    def NeuralNetsRegressor(self):
        # Determine number of nodes per layer through supported
        # functions or provided
        layers = []
        if isinstance(self._mid_num_node_strategy, str) and self._num_layers > 1:
            if self._mid_num_node_strategy == "constant":
                layers = self.constant(
                    self._num_layers, self._start_num_nodes, self._end_num_nodes
                )
            elif self._mid_num_node_strategy == "linear":
                layers = self.linear(
                    self._num_layers, self._start_num_nodes, self._end_num_nodes
                )
            else:
                raise Exception(
                    "Function ("
                    + self._mid_num_node_strategy
                    + ") is not supported, create your own callable"
                )
        else:
            layers = self._mid_num_node_strategy(
                self._num_layers, self._start_num_nodes, self._end_num_nodes
            )

        # Create the deep learning model
        model = Sequential(name=self._name)

        # Initial layer
        model.add(
            Dense(
                units=self._start_num_nodes,
                activation=self._start_activation,
                use_bias=self._start_use_bias,
                kernel_initializer=self._start_kernel_initializer,
                bias_initializer=self._start_bias_initializer,
                kernel_regularizer=self._start_kernel_regularizer,
                bias_regularizer=self._start_bias_regularizer,
                activity_regularizer=self._start_activity_regularizer,
                kernel_constraint=self._start_kernel_constraint,
                bias_constraint=self._start_bias_constraint,
                input_dim=self._input_dim,
            )
        )

        # Dropout
        if self._dropout == True:
            model.add(Dropout(rate=self._rate))

        # Hidden layers
        if self._num_layers > 2:
            for i in range(1, self._num_layers):
                model.add(
                    Dense(
                        units=layers[i],
                        activation=self._mid_activation,
                        use_bias=self._mid_use_bias,
                        kernel_initializer=self._mid_kernel_initializer,
                        bias_initializer=self._mid_bias_initializer,
                        kernel_regularizer=self._mid_kernel_regularizer,
                        bias_regularizer=self._mid_bias_regularizer,
                        activity_regularizer=self._mid_activity_regularizer,
                        kernel_constraint=self._mid_kernel_constraint,
                        bias_constraint=self._mid_bias_constraint,
                    )
                )

        # Ending Layer
        model.add(
            Dense(
                units=self._end_num_nodes,
                activation=self._end_activation,
                use_bias=self._end_use_bias,
                kernel_initializer=self._end_kernel_initializer,
                bias_initializer=self._end_bias_initializer,
                kernel_regularizer=self._end_kernel_regularizer,
                bias_regularizer=self._end_bias_regularizer,
                activity_regularizer=self._end_activity_regularizer,
                kernel_constraint=self._end_kernel_constraint,
                bias_constraint=self._end_bias_constraint,
            )
        )

        # Optimizer
        modelopt = None
        if self._optimizer == "adam":
            modelopt = Adam(
                learning_rate=self._learning_rate,
                beta_1=self._beta_1,
                beta_2=self._beta_2,
                epsilon=self._epsilon,
                amsgrad=self._amsgrad,
                clipnorm=self._clipnorm,
                clipvalue=self._clipvalue,
                global_clipnorm=self._global_clipnorm,
            )
        else:
            raise Exception("Optimizer (" + self._optimizer + ") is not supported")

        # Compile model
        model.compile(
            optimizer=modelopt,
            loss=self._loss,
            metrics=self._metrics,
            loss_weights=self._loss_weights,
            weighted_metrics=self._weighted_metrics,
            run_eagerly=self._run_eagerly,
        )

        # Set model to KerasRegressor from scikeras
        self._model = KerasRegressor(
            model=model,
            random_state=settings.values.random_state,
            warm_start=self._warm_start,
            optimizer=modelopt,
            loss=self._loss,
            metrics=self._metrics,
            batch_size=self._batch_size,
            validation_batch_size=self._validation_batch_size,
            verbose=settings.values.verbosity,
            callbacks=self._callbacks,
            validation_split=self._validation_split,
            shuffle=self._shuffle,
            run_eagerly=self._run_eagerly,
            epochs=self._epochs,
        )

        return self

    # ===========================================================
    # Getters
    @property
    def num_layers(self):
        return self._num_layers

    @property
    def name(self):
        return self._name

    @property
    def dropout(self):
        return self._dropout

    @property
    def rate(self):
        return self._rate

    @property
    def loss(self):
        return self._loss

    @property
    def metrics(self):
        return self._metrics

    @property
    def loss_weights(self):
        return self._loss_weights

    @property
    def run_eagerly(self):
        return self._run_eagerly

    @property
    def steps_per_execution(self):
        return self._steps_per_execution

    @property
    def jit_compile(self):
        return self._jit_compile

    @property
    def batch_size(self):
        return self._batch_size

    @property
    def validation_batch_size(self):
        return self._validation_batch_size

    @property
    def shuffle(self):
        return self._shuffle

    @property
    def callbacks(self):
        return self._callbacks

    @property
    def validation_split(self):
        return self._validation_split

    @property
    def epochs(self):
        return self._epochs

    @property
    def warm_start(self):
        return self._warm_start

    @property
    def start_num_nodes(self):
        return self._start_num_nodes

    @property
    def start_activation(self):
        return self._start_activation

    @property
    def start_use_bias(self):
        return self._start_use_bias

    @property
    def start_kernel_initializer(self):
        return self._start_kernel_initializer

    @property
    def start_bias_initializer(self):
        return self._start_bias_initializer

    @property
    def start_kernel_regularizer(self):
        return self._start_kernel_regularizer

    @property
    def start_bias_regularizer(self):
        return self._start_bias_regularizer

    @property
    def start_activity_regularizer(self):
        return self._start_activity_regularizer

    @property
    def start_kernel_constraint(self):
        return self._start_kernel_constraint

    @property
    def start_bias_constraint(self):
        return self._start_bias_constraint

    @property
    def input_dim(self):
        return self._input_dim

    @property
    def mid_num_node_strategy(self):
        return self._mid_num_node_strategy

    @property
    def mid_activation(self):
        return self._mid_activation

    @property
    def mid_use_bias(self):
        return self._mid_use_bias

    @property
    def mid_kernel_initializer(self):
        return self._mid_kernel_initializer

    @property
    def mid_bias_initializer(self):
        return self._mid_bias_initializer

    @property
    def mid_kernel_regularizer(self):
        return self._mid_kernel_regularizer

    @property
    def mid_bias_regularizer(self):
        return self._mid_bias_regularizer

    @property
    def mid_activity_regularizer(self):
        return self._mid_activity_regularizer

    @property
    def mid_kernel_constraint(self):
        return self._mid_kernel_constraint

    @property
    def mid_bias_constraint(self):
        return self._mid_bias_constraint

    @property
    def end_num_nodes(self):
        return self._end_num_nodes

    @property
    def end_activation(self):
        return self._end_activation

    @property
    def end_use_bias(self):
        return self._end_use_bias

    @property
    def end_kernel_initializer(self):
        return self._end_kernel_initializer

    @property
    def end_bias_initializer(self):
        return self._end_bias_initializer

    @property
    def end_kernel_regularizer(self):
        return self._end_kernel_regularizer

    @property
    def end_bias_regularizer(self):
        return self._end_bias_regularizer

    @property
    def end_activity_regularizer(self):
        return self._end_activity_regularizer

    @property
    def end_kernel_constraint(self):
        return self._start_end_constraint

    @property
    def end_bias_constraint(self):
        return self._end_bias_constraint

    @property
    def optimizer(self):
        return self._optimizer

    @property
    def learning_rate(self):
        return self._learning_rate

    @property
    def beta_1(self):
        return self._beta_1

    @property
    def beta_2(self):
        return self._beta_2

    @property
    def epsilon(self):
        return self._epsilon

    @property
    def amsgrad(self):
        return self._amsgrad

    @property
    def clipnorm(self):
        return self._clipnorm

    @property
    def clipvalue(self):
        return self._clipvalue

    @property
    def global_clipnorm(self):
        return self._global_clipnorm

    # ===========================================================
    # Setters
    @num_layers.setter
    def num_layers(self, num_layers):
        self._num_layers = num_layers

    @name.setter
    def name(self, name):
        self._name = name

    @dropout.setter
    def dropout(self, dropout):
        self._dropout = dropout

    @rate.setter
    def rate(self, rate):
        self._rate = rate

    @loss.setter
    def loss(self, loss):
        self._loss = loss

    @metrics.setter
    def metrics(self, metrics):
        self._metrics = metrics

    @loss_weights.setter
    def loss_weights(self, loss_weights):
        self._loss_weights = loss_weights

    @run_eagerly.setter
    def run_eagerly(self, run_eagerly):
        self._run_eagerly = run_eagerly

    @steps_per_execution.setter
    def steps_per_execution(self, steps_per_execution):
        self._steps_per_execution = steps_per_execution

    @jit_compile.setter
    def jit_compile(self, jit_compile):
        self._jit_compile = jit_compile

    @batch_size.setter
    def batch_size(self, batch_size):
        self._batch_size = batch_size

    @validation_batch_size.setter
    def validation_batch_size(self, validation_batch_size):
        self._validation_batch_size = validation_batch_size

    @shuffle.setter
    def shuffle(self, shuffle):
        self._shuffle = shuffle

    @callbacks.setter
    def callbacks(self, callbacks):
        self._callbacks = callbacks

    @validation_split.setter
    def validation_split(self, validation_split):
        self._validation_split = validation_split

    @epochs.setter
    def epochs(self, epochs):
        self._epochs = epochs

    @warm_start.setter
    def warm_start(self, warm_start):
        self._warm_start = warm_start

    @start_num_nodes.setter
    def start_num_nodes(self, start_num_nodes):
        self._start_num_nodes = start_num_nodes

    @start_activation.setter
    def start_activation(self, start_activation):
        self._start_activation = start_activation

    @start_use_bias.setter
    def start_use_bias(self, start_use_bias):
        self._start_use_bias = start_use_bias

    @start_kernel_initializer.setter
    def start_kernel_initializer(self, start_kernel_initializer):
        self._start_kernel_initializer = start_kernel_initializer

    @start_bias_initializer.setter
    def start_bias_initializer(self, start_bias_initializer):
        self._start_bias_initializer = start_bias_initializer

    @start_kernel_regularizer.setter
    def start_kernel_regularizer(self, start_kernel_regularizer):
        self._start_kernel_regularizer = start_kernel_regularizer

    @start_bias_regularizer.setter
    def start_bias_regularizer(self, start_bias_regularizer):
        self._start_bias_regularizer = start_bias_regularizer

    @start_activity_regularizer.setter
    def start_activity_regularizer(self, start_activity_regularizer):
        self._start_activity_regularizer = start_activity_regularizer

    @start_kernel_constraint.setter
    def start_kernel_constraint(self, start_kernel_constraint):
        self._start_kernel_constraint = start_kernel_constraint

    @start_bias_constraint.setter
    def start_bias_constraint(self, start_bias_constraint):
        self._start_bias_constraint = start_bias_constraint

    @input_dim.setter
    def input_dim(self, input_dim):
        self._input_dim = input_dim

    @mid_num_node_strategy.setter
    def mid_num_node_strategy(self, mid_num_node_strategy):
        assert isinstance(mid_num_node_strategy, str) or isinstance(
            mid_num_node_strategy, callable
        )
        self._mid_num_node_strategy = mid_num_node_strategy

    @mid_activation.setter
    def mid_activation(self, mid_activation):
        self._mid_activation = mid_activation

    @mid_use_bias.setter
    def mid_use_bias(self, mid_use_bias):
        self._mid_use_bias = mid_use_bias

    @mid_kernel_initializer.setter
    def mid_kernel_initializer(self, mid_kernel_initializer):
        self._mid_kernel_initializer = mid_kernel_initializer

    @mid_bias_initializer.setter
    def mid_bias_initializer(self, mid_bias_initializer):
        self._mid_bias_initializer = mid_bias_initializer

    @mid_kernel_regularizer.setter
    def mid_kernel_regularizer(self, mid_kernel_regularizer):
        self._mid_kernel_regularizer = mid_kernel_regularizer

    @mid_bias_regularizer.setter
    def mid_bias_regularizer(self, mid_bias_regularizer):
        self._mid_bias_regularizer = mid_bias_regularizer

    @mid_activity_regularizer.setter
    def mid_activity_regularizer(self, mid_activity_regularizer):
        self._mid_activity_regularizer = mid_activity_regularizer

    @mid_kernel_constraint.setter
    def mid_kernel_constraint(self, mid_kernel_constraint):
        self._mid_kernel_constraint = mid_kernel_constraint

    @mid_bias_constraint.setter
    def mid_bias_constraint(self, mid_bias_constraint):
        self._mid_bias_constraint = mid_bias_constraint

    @end_num_nodes.setter
    def end_num_nodes(self, end_num_nodes):
        self._end_num_nodes = end_num_nodes

    @end_activation.setter
    def end_activation(self, end_activation):
        self._end_activation = end_activation

    @end_use_bias.setter
    def end_use_bias(self, end_use_bias):
        self._end_use_bias = end_use_bias

    @end_kernel_initializer.setter
    def end_kernel_initializer(self, end_kernel_initializer):
        self._end_kernel_initializer = end_kernel_initializer

    @end_bias_initializer.setter
    def end_bias_initializer(self, end_bias_initializer):
        self._end_bias_initializer = end_bias_initializer

    @end_kernel_regularizer.setter
    def end_kernel_regularizer(self, end_kernel_regularizer):
        self._end_kernel_regularizer = end_kernel_regularizer

    @end_bias_regularizer.setter
    def end_bias_regularizer(self, end_bias_regularizer):
        self._end_bias_regularizer = end_bias_regularizer

    @end_activity_regularizer.setter
    def end_activity_regularizer(self, end_activity_regularizer):
        self._end_activity_regularizer = end_activity_regularizer

    @end_kernel_constraint.setter
    def end_kernel_constraint(self, end_kernel_constraint):
        self._start_end_constraint = end_kernel_constraint

    @end_bias_constraint.setter
    def end_bias_constraint(self, end_bias_constraint):
        self._end_bias_constraint = end_bias_constraint

    @optimizer.setter
    def optimizer(self, optimizer):
        self._optimizer = optimizer

    @learning_rate.setter
    def learning_rate(self, learning_rate):
        self._learning_rate = learning_rate

    @beta_1.setter
    def beta_1(self, beta_1):
        self._beta_1 = beta_1

    @beta_2.setter
    def beta_2(self, beta_2):
        self._beta_2 = beta_2

    @epsilon.setter
    def epsilon(self, epsilon):
        self._epsilon = epsilon

    @amsgrad.setter
    def amsgrad(self, amsgrad):
        self._amsgrad = amsgrad

    @clipnorm.setter
    def clipnorm(self, clipnorm):
        self._clipnorm = clipnorm

    @clipvalue.setter
    def clipvalue(self, clipvalue):
        self._clipvalue = clipvalue

    @global_clipnorm.setter
    def global_clipnorm(self, global_clipnorm):
        self._global_clipnorm = global_clipnorm



import pyMAISE.settings as settings

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler, StandardScaler


class PreProcesser:
    def __init__(
        self,
        path,
        input_slice: slice = None,
        output_slice: slice = None,
        header="infer",
    ):
        if isinstance(path, str):
            # Read in data from path
            self._data = pd.read_csv(path, header=header)
            self._input_slice = input_slice
            self._output_slice = output_slice

            # Separate data based on x and y indicies
            self._inputs = self._data.iloc[:, input_slice]
            self._outputs = self._data.iloc[:, output_slice]
        elif isinstance(path, list):
            # Make sure only two paths are given
            assert len(path) == 2

            # Assuming the first path is the input and second is
            # output data
            self._inputs = pd.read_csv(path[0], header=header)
            self._outputs = pd.read_csv(path[1], header=header)

            # Join input and output DataFrames
            self._data = pd.concat([self._inputs, self._outputs], axis=1)

            # Fill slicing information
            self._input_slice = slice(0, self._inputs.shape[1])
            self._output_slice = slice(
                self._inputs.shape[1] - 1, self._outputs.shape[1]
            )

        # Scalers
        self._xscaler = None
        self._yscaler = None

        if settings.values.verbosity > 0:
            print("Reading data from", path)

            if settings.values.verbosity == 2:
                print(self._data.head(5))

    def data_split(self):
        return train_test_split(
            self._inputs,
            self._outputs,
            test_size=settings.values.test_size,
            random_state=settings.values.random_state,
        )

    def min_max_scale(self, scale_x: bool = True, scale_y: bool = True):
        # Split data
        xtrain, xtest, ytrain, ytest = self.data_split()

        # Scale x
        if scale_x:
            self._xscaler = MinMaxScaler()
            xtrain = pd.DataFrame(
                self._xscaler.fit_transform(xtrain),
                index=xtrain.index,
                columns=xtrain.columns,
            )
            xtest = pd.DataFrame(
                self._xscaler.transform(xtest),
                index=xtest.index,
                columns=xtest.columns,
            )

        # Scale y
        if scale_y:
            self._yscaler = MinMaxScaler()
            ytrain = pd.DataFrame(
                self._yscaler.fit_transform(ytrain),
                index=ytrain.index,
                columns=ytrain.columns,
            )
            ytest = pd.DataFrame(
                self._yscaler.transform(ytest),
                index=ytest.index,
                columns=ytest.columns,
            )

        if settings.values.verbosity > 0:
            print("MinMax scaling data")

        return (xtrain, xtest, ytrain, ytest)

    def std_scale(self, scale_x: bool = True, scale_y: bool = True):
        # Split data
        xtrain, xtest, ytrain, ytest = self.data_split()

        # Scale x
        if scale_x:
            self._xscaler = StandardScaler()
            xtrain = pd.DataFrame(
                self._xscaler.fit_transform(xtrain),
                index=xtrain.index,
                columns=xtrain.columns,
            )
            xtest = pd.DataFrame(
                self._xscaler.transform(xtest),
                index=xtest.index,
                columns=xtest.columns,
            )

        # Scale y
        if scale_y:
            self._yscaler = StandardScaler()
            ytrain = pd.DataFrame(
                self._yscaler.fit_transform(ytrain),
                index=ytrain.index,
                columns=ytrain.columns,
            )
            ytest = pd.DataFrame(
                self._yscaler.transform(ytest),
                index=ytest.index,
                columns=ytest.columns,
            )

        if settings.values.verbosity > 0:
            print("MinMax scaling data")

        return (xtrain, xtest, ytrain, ytest)

    def correlation_matrix(
        self,
        method="pearson",
        min_periods=1,
        fig=None,
        ax=None,
        colorbar=True,
        annotations=False,
    ):
        # Create correlation matrix DataFrame
        corr = self._data.corr(method, min_periods)

        # Get figure and axis object if not provided
        if ax == None and fig == None:
            fig, ax = plt.subplots()

        # Create heatmap
        im = ax.imshow(corr)

        # Set x and y ticks to column headers
        ax.set_xticks(np.arange(corr.shape[0]), labels=corr.columns, rotation=65)
        ax.set_yticks(np.arange(corr.shape[1]), labels=corr.columns, rotation=0)

        # Add colorbar if colorbar == True
        if colorbar:
            fig.colorbar(im)

        # Add annotations of each value in square if annotations == True
        if annotations:
            for i in range(len(corr.columns)):
                for j in range(len(corr.columns)):
                    ax.text(
                        j,
                        i,
                        round(corr.to_numpy()[i, j], 2),
                        ha="center",
                        va="center",
                        color="black",
                    )

        return (fig, ax)

    # Getters
    @property
    def data(self) -> pd.DataFrame:
        return self._data

    @property
    def inputs(self) -> pd.DataFrame:
        return self._inputs

    @property
    def outputs(self) -> pd.DataFrame:
        return self._outputs

    @property
    def xscaler(self):
        return self._xscaler

    @property
    def yscaler(self):
        return self._yscaler



import pyMAISE.settings as settings

from sklearn.tree import DecisionTreeRegressor


class DecisionTreeRegression:
    def __init__(self, parameters: dict = None):
        # Model parameters
        self._criterion = "squared_error"
        self._splitter = "best"
        self._max_depth = None
        self._min_samples_split = 2
        self._min_samples_leaf = 1
        self._min_weight_fraction_leaf = 0.0
        self._max_features = None
        self._max_leaf_nodes = None
        self._min_impurity_decrease = 0.0
        self._ccp_alpha = 0.0

        # Change if user provided changes in dictionary
        if parameters != None:
            for key, value in parameters.items():
                setattr(self, key, value)

    # ===========================================================
    # Methods
    def regressor(self):
        return DecisionTreeRegressor(
            criterion=self._criterion,
            splitter=self._splitter,
            max_depth=self._max_depth,
            min_samples_split=self._min_samples_split,
            min_samples_leaf=self._min_samples_leaf,
            min_weight_fraction_leaf=self._min_weight_fraction_leaf,
            max_features=self._max_features,
            random_state=settings.values.random_state,
            max_leaf_nodes=self._max_leaf_nodes,
            min_impurity_decrease=self._min_impurity_decrease,
            ccp_alpha=self._ccp_alpha,
        )

    # ===========================================================
    # Getters
    @property
    def criterion(self) -> str:
        return self._criterion

    @property
    def splitter(self) -> str:
        return self._splitter

    @property
    def max_depth(self) -> int:
        return self.max_depth

    @property
    def min_samples_split(self) -> float:
        return self._min_samples_split

    @property
    def min_samples_leaf(self) -> float:
        return self._min_samples_leaf

    @property
    def min_weight_fraction_leaf(self) -> float:
        return self._min_weight_fraction_leaf

    @property
    def max_features(self):
        return self._max_features

    @property
    def max_leaf_nodes(self) -> int:
        return self._max_leaf_nodes

    @property
    def min_impurity_decrease(self) -> float:
        return self._min_impurity_decrease

    @property
    def ccp_alpha(self) -> float:
        return self._ccp_alpha

    # ===========================================================
    # Setters
    @criterion.setter
    def criterion(self, criterion: str):
        assert (
            criterion == "squared_error"
            or criterion == "friedman_mse"
            or criterion == "absolute_error"
            or criterion == "poisson"
        )
        self._criterion = criterion

    @splitter.setter
    def splitter(self, splitter: str):
        assert splitter == "best" or splitter == "random"
        self._splitter = splitter

    @max_depth.setter
    def max_depth(self, max_depth: int):
        assert max_depth > 1 or max_depth == None
        self._max_depth = max_depth

    @min_samples_split.setter
    def min_samples_split(self, min_samples_split: float):
        assert min_samples_split >= 2
        self._min_samples_split = min_samples_split

    @min_samples_leaf.setter
    def min_samples_leaf(self, min_samples_leaf: float):
        assert min_samples_leaf >= 1
        self._min_samples_leaf = min_samples_leaf

    @min_weight_fraction_leaf.setter
    def min_weight_fraction_leaf(self, min_weight_fraction_leaf: float):
        assert min_weight_fraction_leaf >= 0
        self._min_weight_fraction_leaf = min_weight_fraction_leaf

    @max_features.setter
    def max_features(self, max_features):
        self._max_features = max_features

    @max_leaf_nodes.setter
    def max_leaf_nodes(self, max_leaf_nodes: int):
        assert max_leaf_nodes >= 0 or max_leaf_nodes == None
        self._max_leaf_nodes = max_leaf_nodes

    @min_impurity_decrease.setter
    def min_impurity_decrease(self, min_impurity_decrease: float):
        self._min_impurity_decrease = min_impurity_decrease

    @ccp_alpha.setter
    def ccp_alpha(self, ccp_alpha: float):
        assert ccp_alpha >= 0
        self._ccp_alpha = ccp_alpha



from sklearn.neighbors import KNeighborsRegressor


class KNeighborsRegression:
    def __init__(self, parameters: dict = None):
        # Model Parameters
        self._n_neighbors = 5
        self._weights = "uniform"
        self._algorithm = "auto"
        self._leaf_size = 30
        self._p = 2
        self._metric = "minkowski"
        self._metric_params = None
        self._n_jobs = None

        # Change if user provided changes in dictionary
        if parameters != None:
            for key, value in parameters.items():
                setattr(self, key, value)

    # ===========================================================
    # Methods
    def regressor(self):
        return KNeighborsRegressor(
            n_neighbors=self._n_neighbors,
            weights=self._weights,
            algorithm=self._algorithm,
            leaf_size=self._leaf_size,
            p=self._p,
            metric=self._metric,
            metric_params=self._metric_params,
            n_jobs=self._n_jobs,
        )

    # ===========================================================
    # Getters
    @property
    def n_neighbors(self) -> int:
        return self._n_neighbors

    @property
    def weights(self):
        return self._weights

    @property
    def algorithm(self) -> str:
        return self._algorithm

    @property
    def leaf_size(self) -> int:
        return self._leaf_size

    @property
    def p(self) -> int:
        return self._p

    @property
    def metric(self):
        return self._metric

    @property
    def metric_params(self) -> dict:
        return self._metric_params

    @property
    def n_jobs(self) -> int:
        return self._n_jobs

    # ===========================================================
    # Setters
    @n_neighbors.setter
    def n_neighbors(self, n_neighbors: int):
        assert n_neighbors >= 1
        self._n_neighbors = n_neighbors

    @weights.setter
    def weights(self, weights):
        self._weights = weights

    @algorithm.setter
    def algorithm(self, algorithm: str):
        assert (
            algorithm == "auto"
            or algorithm == "ball_tree"
            or algorithm == "kd_tree"
            or algorithm == "brute"
        )
        self._algorithm = algorithm

    @leaf_size.setter
    def leaf_size(self, leaf_size: int):
        assert leaf_size > 0
        self._leaf_size = leaf_size

    @p.setter
    def p(self, p: int):
        assert p > 0
        self._p = p

    @metric.setter
    def metric(self, metric):
        self._metric = metric

    @metric_params.setter
    def metric_params(self, metric_params: dict):
        assert metric_params == None or isinstance(metric_params, dict)
        self._metric_params = metric_params

    @n_jobs.setter
    def n_jobs(self, n_jobs: int):
        assert n_jobs == None or n_jobs >= -1
        self._n_jobs = n_jobs



import pyMAISE.settings as settings
from pyMAISE.methods import *

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
from skopt.space import Integer, Real, Categorical
from skopt import BayesSearchCV


class Tuning:
    def __init__(self, data: pd.DataFrame, model_settings: dict):
        # Extract training data
        self._xtrain = data[0]
        self._ytrain = data[2]

        # Extract target models from dictionary
        models_str = model_settings["models"]

        # Tuning loss for convergence plots
        self._tuning = {}

        # Initialize models of interest
        self._models = {}
        for model in models_str:
            # Pull parameters provided
            parameters = None
            if model in model_settings:
                parameters = model_settings[model]

            # Add model object to dictionary
            if model == "linear":
                self._models[model] = LinearRegression(parameters=parameters)
            elif model == "lasso":
                self._models[model] = LassoRegression(parameters=parameters)
            elif model == "svr":
                if self._ytrain.shape[1] > 1:
                    raise Exception("SVR does not support multi-output data sets")
                self._models[model] = SVRegression(parameters=parameters)
            elif model == "dtree":
                self._models[model] = DecisionTreeRegression(parameters=parameters)
            elif model == "rforest":
                self._models[model] = RandomForestRegression(parameters=parameters)
            elif model == "knn":
                self._models[model] = KNeighborsRegression(parameters=parameters)
            elif model == "nn":
                self._models[model] = NeuralNetsRegression(parameters=parameters)
            else:
                raise Exception(
                    "The model requested ("
                    + model
                    + ") is either misspelled or not supported"
                )

        # For single input or outputs the data must be converted to a 1D array to
        # get rid of sklearn warnings
        if self._xtrain.shape[1] == 1:
            self._xtrain = self._xtrain.iloc[:, 0]
        if self._ytrain.shape[1] == 1:
            self._ytrain = self._ytrain.iloc[:, 0]

    # ===========================================================
    # Methods
    def manual_search(self, models: list = None, model_settings=None):
        # Get model types if not provided
        if models == None:
            models = list(self._models.keys())

        data = {}
        for model in models:
            if settings.values.verbosity > 0:
                print("-- " + model)

            # Run model
            estimator = self._models[model].regressor()
            if model_settings != None and model in model_settings:
                estimator.set_params(model_settings)

            resulting_model = estimator.fit(self._xtrain, self._ytrain)

            # Place model in
            data[model] = (
                pd.DataFrame({"params": [resulting_model.get_params()]}),
                resulting_model,
            )

        return data

    def grid_search(
        self,
        param_spaces: dict,
        scoring=None,
        models: list = None,
        n_jobs: int = None,
        refit=True,
        cv=None,
        pre_dispatch="2*n_jobs",
    ):
        if models == None:
            models = list(self._models.keys())

        if settings.values.verbosity > 0:
            print("Hyper-parameter tuning with grid search")

        data = {}
        for model in models:
            if model in param_spaces:
                if settings.values.verbosity > 0:
                    print("-- " + model)

                # Run grid search
                search = GridSearchCV(
                    estimator=self._models[model].regressor(),
                    param_grid=param_spaces[model],
                    scoring=scoring,
                    n_jobs=n_jobs,
                    refit=refit,
                    cv=cv,
                    verbose=settings.values.verbosity,
                    pre_dispatch=pre_dispatch,
                )
                resulting_models = search.fit(self._xtrain, self._ytrain)

                # Add tuning results for convergence plot
                cv_results = pd.DataFrame(resulting_models.cv_results_)
                self._tuning[model] = cv_results["mean_test_score"]

                # Place parameter configurations in DataFrame and sort based on rank,
                # save the top num_configs_saved to the data dictionary
                # Exclude timing data from DataFrame
                top_configs = cv_results.sort_values("rank_test_score").iloc[
                    : settings.values.num_configs_saved, :
                ]

                if settings.values.verbosity > 1:
                    print(top_configs)

                data[model] = (top_configs, resulting_models.best_estimator_)

            else:
                print(
                    "Hyper-parameter tuning search space was not provided for "
                    + model
                    + ", doing manual fit"
                )
                data = {**data, **self.manual_search(models=[model])}

        return data

    def random_search(
        self,
        param_spaces: dict,
        scoring=None,
        models: list = None,
        n_iter: int = 10,
        n_jobs: int = None,
        refit=True,
        cv=None,
        pre_dispatch="2*n_jobs",
    ):
        if models == None:
            models = list(self._models.keys())

        if settings.values.verbosity > 0:
            print("Hyper-parameter tuning with random search")

        data = {}

        for model in models:
            if model in param_spaces:
                if settings.values.verbosity > 0:
                    print("-- " + model)

                # Run random search
                search = RandomizedSearchCV(
                    estimator=self._models[model].regressor(),
                    param_distributions=param_spaces[model],
                    scoring=scoring,
                    n_iter=n_iter,
                    n_jobs=n_jobs,
                    refit=refit,
                    cv=cv,
                    verbose=settings.values.verbosity,
                    random_state=settings.values.random_state,
                    pre_dispatch=pre_dispatch,
                )
                resulting_models = search.fit(self._xtrain, self._ytrain)

                # Add tuning results for convergence plot
                cv_results = pd.DataFrame(resulting_models.cv_results_)
                self._tuning[model] = cv_results["mean_test_score"]

                # Place parameter configurations in DataFrame and sort based on rank,
                # save the top num_configs_saved to the data dictionary
                # Exclude timing data from DataFrame
                top_configs = cv_results.sort_values("rank_test_score").iloc[
                    : settings.values.num_configs_saved, :
                ]

                if settings.values.verbosity > 1:
                    print(top_configs)

                data[model] = (top_configs, resulting_models.best_estimator_)

            else:
                print(
                    "Hyper-parameter tuning search space was not provided for "
                    + model
                    + ", doing manual fit"
                )
                data = {**data, **self.manual_search(models=[model])}

        return data

    def bayesian_search(
        self,
        param_spaces: dict,
        scoring=None,
        models: list = None,
        n_iter: int = 50,
        optimizer_kwargs: dict = None,
        fit_params: dict = None,
        n_jobs: int = None,
        n_points: int = 1,
        refit=True,
        cv=None,
        pre_dispatch="2*n_jobs",
    ):
        if models == None:
            models = list(self._models.keys())

        if settings.values.verbosity > 0:
            print("Hyper-parameter tuning with bayesian search")

        data = {}
        for model in models:
            if model in param_spaces:
                if settings.values.verbosity > 0:
                    print("-- " + model)

                # Convert list of values to search space dimensions
                for key, value in param_spaces[model].items():
                    if isinstance(value[0], int):
                        param_spaces[model][key] = Integer(
                            low=np.min(value), high=np.max(value), name=key
                        )
                    elif isinstance(value[0], float):
                        param_spaces[model][key] = Real(
                            low=np.min(value), high=np.max(value), name=key
                        )
                    elif isinstance(value[0], str):
                        param_spaces[model][key] = Categorical(
                            categories=value, name=key
                        )

                # Run Bayesian search
                search = BayesSearchCV(
                    estimator=self._models[model].regressor(),
                    search_spaces=param_spaces[model],
                    n_iter=n_iter,
                    optimizer_kwargs=optimizer_kwargs,
                    scoring=scoring,
                    fit_params=fit_params,
                    n_jobs=n_jobs,
                    n_points=n_points,
                    pre_dispatch=pre_dispatch,
                    cv=cv,
                    refit=refit,
                    verbose=settings.values.verbosity,
                    random_state=settings.values.random_state,
                )
                resulting_models = search.fit(self._xtrain, self._ytrain)

                # Add tuning results for convergence plot
                cv_results = pd.DataFrame(resulting_models.cv_results_)
                self._tuning[model] = cv_results["mean_test_score"]

                # Place parameter configurations in DataFrame and sort based on rank,
                # save the top num_configs_saved to the data dictionary
                # Exclude timing data from DataFrame
                top_configs = cv_results.sort_values("rank_test_score").iloc[
                    : settings.values.num_configs_saved, :
                ]

                if settings.values.verbosity > 1:
                    print(top_configs)

                data[model] = (top_configs, resulting_models.best_estimator_)

            else:
                print(
                    "Hyper-parameter tuning search space was not provided for "
                    + model
                    + ", doing manual fit"
                )
                data = {**data, **self.manual_search(models=[model])}

        return data

    def convergence_plot(self, ax=None, model_types=None):
        # If no models are provided fit all
        if model_types == None:
            model_types = self._tuning.keys()
        elif isinstance(model_types, str):
            model_types = [model_types]

        if ax == None:
            ax = plt.gca()

        for model in model_types:
            ax.plot(
                np.linspace(
                    1, self._tuning[model].shape[0], self._tuning[model].shape[0]
                ),
                self._tuning[model],
                linestyle="-",
                marker="o",
                label=model,
            )

        # Show legend if length of models is more than one
        if len(model_types) > 1:
            ax.legend()

        ax.set_xlabel("Iteration")
        ax.set_ylabel("Mean Test Score")

        return ax



from pyMAISE.methods import *

import math
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from sklearn.metrics import r2_score, mean_absolute_error, mean_squared_error


class PostProcessor:
    def __init__(
        self,
        data: pd.DataFrame,
        models_list,
        new_model_settings: dict = None,
        yscaler=None,
    ):
        # Extract data
        self._xtrain = data[0]
        self._xtest = data[1]
        self._ytrain = data[2]
        self._ytest = data[3]

        # Initialize models of interest
        model_types = []
        params = []
        model_wrappers = []

        if isinstance(models_list, dict):
            models_list = [models_list]

        # Extract models and the DataFrame of hyper-parameter configurations
        for models in models_list:
            for model, configs in models.items():
                # Fill model types list with string of model type
                model_types = model_types + [model] * configs[0].shape[0]

                # Fil parameter string with hyper-parameter configurations for each type
                params = params + configs[0]["params"].tolist()

                # Get all model wrappers and update parameter configurations if needed
                estimator = configs[1]
                if new_model_settings != None and model in new_model_settings:
                    estimator = estimator.set_params(**new_model_settings[model])
                model_wrappers = model_wrappers + [estimator] * configs[0].shape[0]

        # Create models DataFrame
        self._models = pd.DataFrame(
            {
                "Model Types": model_types,
                "Parameter Configurations": params,
                "Model Wrappers": model_wrappers,
            }
        )

        # Fit each model to training data and get predicted training
        # and testing from each model
        yhat_train, yhat_test, histories = self._fit()

        if yscaler != None:
            self._yscaler = yscaler
            self._ytrain = pd.DataFrame(
                yscaler.inverse_transform(self._ytrain),
                index=self._ytrain.index,
                columns=self._ytrain.columns,
            )
            self._ytest = pd.DataFrame(
                yscaler.inverse_transform(self._ytest),
                index=self._ytest.index,
                columns=self._ytest.columns,
            )

            for i in range(len(yhat_train)):
                yhat_train[i] = yscaler.inverse_transform(yhat_train[i])
                yhat_test[i] = yscaler.inverse_transform(yhat_test[i])

        self._models = pd.concat(
            [
                self._models,
                pd.DataFrame(
                    {
                        "Train Yhat": yhat_train,
                        "Test Yhat": yhat_test,
                        "History": histories,
                    }
                ),
            ],
            axis=1,
        )

    # ===========================================================
    # Methods
    def _fit(self):
        # Array for trainig and testing prediceted outcomes
        yhat_train = []
        yhat_test = []
        histories = []

        ytrain = self._ytrain
        if ytrain.shape[1] == 1:
            ytrain = self._ytrain.values.ravel()

        # Fit each model and predict outcomes
        for i in range(self._models.shape[0]):
            # Estract regressor for the configuration
            regressor = self._models["Model Wrappers"][i].set_params(
                **self._models["Parameter Configurations"][i]
            )

            # Append learning curve history of neural networks and run fit for all
            if self._models["Model Types"][i] != "nn":
                regressor.fit(self._xtrain, ytrain)
                histories.append(None)
            else:
                histories.append(
                    regressor.fit(self._xtrain, ytrain).model.history.history
                )

            # Append training and testing predictions
            yhat_train.append(
                regressor.predict(self._xtrain).reshape(self._ytrain.shape)
            )
            yhat_test.append(regressor.predict(self._xtest).reshape(self._ytest.shape))

        return (yhat_train, yhat_test, histories)

    def metrics(self, sort_by="Test R2", model_type: str = None, y=None):
        # Get the list of y if not provided
        if y == None:
            y = slice(0, len(self._ytrain.columns) + 1)
        elif isinstance(y, str):
            y = self._ytrain.columns.get_loc(y)

        train_r2 = []
        train_mae = []
        train_mse = []
        train_mse = []
        train_rmse = []
        test_r2 = []
        test_mae = []
        test_mse = []
        test_mse = []
        test_rmse = []

        for i in range(self._models.shape[0]):
            # Get predictions
            yhat_train = self._models["Train Yhat"][i]
            yhat_test = self._models["Test Yhat"][i]

            # Calculate performance metrics and append to lists
            train_r2.append(
                r2_score(self._ytrain[self._ytrain.columns[y]], yhat_train[:, y])
            )
            train_mae.append(
                mean_absolute_error(
                    self._ytrain[self._ytrain.columns[y]], yhat_train[:, y]
                )
            )
            train_mse.append(
                mean_squared_error(
                    self._ytrain[self._ytrain.columns[y]], yhat_train[:, y]
                )
            )
            train_rmse.append(math.sqrt(train_mse[i]))
            test_r2.append(
                r2_score(self._ytest[self._ytest.columns[y]], yhat_test[:, y])
            )
            test_mae.append(
                mean_absolute_error(
                    self._ytest[self._ytest.columns[y]], yhat_test[:, y]
                )
            )
            test_mse.append(
                mean_squared_error(self._ytest[self._ytest.columns[y]], yhat_test[:, y])
            )
            test_rmse.append(math.sqrt(test_mse[i]))

        # If the sort_by is anything other than R2, metrics is ascending
        ascending = True
        if sort_by == "Test R2" or sort_by == "Train R2":
            ascending = False

        self._models["Train R2"] = train_r2
        self._models["Train MAE"] = train_mae
        self._models["Train MSE"] = train_mse
        self._models["Train RMSE"] = train_rmse
        self._models["Test R2"] = test_r2
        self._models["Test MAE"] = test_mae
        self._models["Test MSE"] = test_mse
        self._models["Test RMSE"] = test_rmse

        models = self._models[
            [
                "Model Types",
                "Parameter Configurations",
                "Train R2",
                "Train MAE",
                "Train MSE",
                "Train RMSE",
                "Test R2",
                "Test MAE",
                "Test MSE",
                "Test RMSE",
            ]
        ]

        if model_type == None:
            return models.sort_values(sort_by, ascending=[ascending])
        else:
            return models[models["Model Types"] == model_type].sort_values(
                sort_by, ascending=[ascending]
            )

    def _get_idx(self, idx: int = None, model_type: str = None, sort_by="Test R2"):
        # Determine the index of the model in the DataFrame
        if idx == None:
            if model_type != None:
                # If an index is not given but a model type is, get index based on sort_by
                if sort_by == "Test R2" or sort_by == "Train R2":
                    idx = self._models[self._models["Model Types"] == model_type][
                        sort_by
                    ].idxmax()
                else:
                    idx = self._models[self._models["Model Types"] == model_type][
                        sort_by
                    ].idxmin()
            else:
                # If an index is not given and the model type is not given, return index
                # of best in sort_by
                if sort_by == "Test R2" or sort_by == "Train R2":
                    idx = self._models[sort_by].idxmax()
                else:
                    idx = self._models[sort_by].idxmin()

        return idx

    def get_predictions(
        self, idx: int = None, model_type: str = None, sort_by="Test R2"
    ):
        # Determine the index of the model in the DataFrame
        idx = self._get_idx(idx=idx, model_type=model_type, sort_by=sort_by)

        yhat_train = self._models["Train Yhat"][idx]
        yhat_test = self._models["Test Yhat"][idx]

        return (yhat_train, yhat_test)

    def get_params(
        self,
        idx: int = None,
        model_type: str = None,
        sort_by="Test R2",
        full: bool = False,
    ):
        # Determine the index of the model in the DataFrame
        idx = self._get_idx(idx=idx, model_type=model_type, sort_by=sort_by)

        parameters = self._models["Parameter Configurations"][idx]
        model_type = self._models["Model Types"][idx]

        if full == True:
            estimator = self._models["Model Wrappers"][idx].set_params(**parameters)
            parameters = estimator.get_params()

        return pd.DataFrame({"Model Types": [model_type], **parameters})

    def get_model(self, idx: int = None, model_type: str = None, sort_by="Test R2"):
        # Determine the index of the model in the DataFrame
        idx = self._get_idx(idx=idx, model_type=model_type, sort_by=sort_by)

        if self._yscaler:
            ytrain = self._yscaler.transform(self._ytrain)

        # Get regressor and fit the model
        regressor = self._models["Model Wrappers"][idx].set_params(
            **self._models["Parameter Configurations"][idx]
        )
        regressor.fit(self._xtrain, ytrain)

        return regressor

    def diagonal_validation_plot(
        self,
        ax=None,
        idx: int = None,
        model_type: str = None,
        sort_by="Test R2",
        y: list = None,
    ):
        # Determine the index of the model in the DataFrame
        idx = self._get_idx(idx=idx, model_type=model_type, sort_by=sort_by)

        # Get the list of y if not provided
        if y == None:
            y = list(range(self._ytrain.shape[1]))

        # Get prediected and actual outputs
        yhat_train = self._models["Train Yhat"][idx]
        yhat_test = self._models["Test Yhat"][idx]
        ytrain = self._ytrain.to_numpy()
        ytest = self._ytest.to_numpy()

        if ax == None:
            ax = plt.gca()

        for y_idx in y:
            if isinstance(y_idx, str):
                y_idx = self._ytest.columns.get_loc(y_idx)

            ax.scatter(yhat_train[:, y_idx], ytrain[:, y_idx], c="b", marker="o")
            ax.scatter(yhat_test[:, y_idx], ytest[:, y_idx], c="r", marker="o")

        lims = [
            np.min([ax.get_xlim(), ax.get_ylim()]),
            np.max([ax.get_xlim(), ax.get_ylim()]),
        ]

        ax.plot(lims, lims, "k--")
        ax.set_aspect("equal")
        ax.set_xlim(lims)
        ax.set_ylim(lims)
        ax.legend(["Training Data", "Testing Data"])
        ax.set_xlabel("Predicted Outcome")
        ax.set_ylabel("Actual Outcome")

        return ax

    def validation_plot(
        self,
        ax=None,
        idx: int = None,
        model_type: str = None,
        sort_by="Test R2",
        y: list = None,
    ):
        # Determine the index of the model in the DataFrame
        idx = self._get_idx(idx=idx, model_type=model_type, sort_by=sort_by)

        # Get the list of y if not provided
        if y == None:
            y = list(range(self._ytest.shape[1]))

        # Get prediected and actual outputs
        yhat_test = self._models["Test Yhat"][idx]
        ytest = self._ytest.to_numpy()

        if ax == None:
            ax = plt.gca()

        for y_idx in y:
            # If the column name is given as opposed to the position,
            # find the position
            if isinstance(y_idx, str):
                y_idx = self._ytest.columns.get_loc(y_idx)

            ax.plot(
                np.linspace(1, ytest.shape[0], ytest.shape[0]),
                np.abs((ytest[:, y_idx] - yhat_test[:, y_idx]) / ytest[:, y_idx]) * 100,
                "-o",
                label=self._ytest.columns[y_idx],
            )

        if len(y) > 1:
            ax.legend()

        ax.set_xlabel("Testing Data Index")
        ax.set_ylabel("Absolute Relative Error (%)")

        return ax

    def nn_learning_plot(
        self,
        ax=None,
        idx: int = None,
        sort_by="Test R2",
    ):
        # Determine the index of the model in the DataFrame
        idx = self._get_idx(idx=idx, model_type="nn", sort_by=sort_by)

        if ax == None:
            ax = plt.gca()

        history = self._models["History"][idx]

        ax.plot(history["loss"], label="Training")
        ax.plot(history["val_loss"], label="Validation")
        ax.legend()
        ax.set_xlabel("Epoch")
        ax.set_ylabel("Loss")

        return ax



from setuptools import setup, find_packages

VERSION = "0.0.1"
DESCRIPTION = "Michigan Artificial Intelligance Standard Environment"
LONG_DESCRIPTION = "Machine learning and artificial intelligance benchmarking library for nuclear engineering applications."

setup(
    name="pyMAISE",
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(include=["pyMAISE", "pyMAISE.*"]),
    install_requires=[
        "pandas",
        "numpy",
        "scikit-learn",
        "scikit-optimize",
        "keras>=2.12.0",
        "tensorflow>=2.12.0",
        "scikeras",
        "matplotlib",
    ],
    extras_require={"test": ["pytest"], "stats": ["scipy"]},
    package_data={"pyMAISE.data": ["*.csv"]},
    author="Patrick Myers",
    author_email="myerspat@umich.edu",
)



import pyMAISE.settings as settings

from sklearn.linear_model import Lasso


class LassoRegression:
    def __init__(self, parameters: dict = None):
        # Model Parameters
        self._alpha = 1.0
        self._fit_intercept = True
        self._precompute = False
        self._copy_X = True
        self._max_iter = 1000
        self._tol = 1e-4
        self._warm_start = False
        self._positive = False
        self._selection = "cyclic"

        # Change if user provided changes in dictionary
        if parameters != None:
            for key, value in parameters.items():
                setattr(self, key, value)

    # ===========================================================
    # Methods
    def regressor(self):
        return Lasso(
            alpha=self._alpha,
            fit_intercept=self._fit_intercept,
            precompute=self._precompute,
            copy_X=self._copy_X,
            max_iter=self._max_iter,
            tol=self._tol,
            warm_start=self._warm_start,
            positive=self._positive,
            selection=self._selection,
            random_state=settings.values.random_state,
        )

    # ===========================================================
    # Getters
    @property
    def alpha(self) -> float:
        return self._alpha

    @property
    def fit_intercept(self) -> bool:
        return self._fit_intercept

    @property
    def precompute(self) -> bool:
        return self._precompute

    @property
    def copy_X(self) -> bool:
        return self._copy_X

    @property
    def max_iter(self) -> int:
        return self._max_iter

    @property
    def positive(self) -> bool:
        return self._positive

    @property
    def selection(self) -> str:
        return self._selection

    # ===========================================================
    # Setters
    @alpha.setter
    def alpha(self, alpha: float):
        assert alpha >= 0
        self._alpha = alpha

    @fit_intercept.setter
    def fit_intercept(self, fit_intercept: bool):
        self._fit_intercept = fit_intercept

    @precompute.setter
    def precompute(self, precompute: bool):
        self._precompute = precompute

    @copy_X.setter
    def copy_X(self, copy_X: bool):
        self._copy_X = copy_X

    @max_iter.setter
    def max_iter(self, max_iter: int):
        assert max_iter > 0
        self._max_iter = max_iter

    @positive.setter
    def positive(self, positive: bool):
        self._positive = positive

    @selection.setter
    def selection(self, selection: str):
        assert selection == "cyclic" or selection == "random"
        self._selection = selection



import pyMAISE.settings as settings

from sklearn.svm import SVR


class SVRegression:
    def __init__(self, parameters: dict = None):
        # Model parameters
        self._kernel = "rbf"
        self._degree = 3
        self._gamma = "scale"
        self._coef0 = 0.0
        self._tol = 1e-3
        self._C = 1.0
        self._epsilon = 0.1
        self._shrinking = True
        self._cache_size = 200
        self._max_iter = -1

        # Change if user provided changes in dictionary
        if parameters != None:
            for key, value in parameters.items():
                setattr(self, key, value)

    # ===========================================================
    # Methods
    def regressor(self):
        return SVR(
            kernel=self._kernel,
            degree=self._degree,
            gamma=self._gamma,
            coef0=self._coef0,
            tol=self._tol,
            C=self._C,
            epsilon=self._epsilon,
            shrinking=self._shrinking,
            cache_size=self._cache_size,
            verbose=settings.values.verbosity,
            max_iter=self._max_iter,
        )

    # ===========================================================
    # Getters
    @property
    def kernel(self):
        return self._kernel

    @property
    def degree(self) -> int:
        return self._degree

    @property
    def gamma(self):
        return self.gamma

    @property
    def coef0(self) -> float:
        return self._coef0

    @property
    def tol(self) -> float:
        return self._tol

    @property
    def C(self) -> float:
        return self._C

    @property
    def epsilon(self) -> float:
        return self._epsilon

    @property
    def shrinking(self) -> bool:
        return self._shrinking

    @property
    def cache_size(self) -> float:
        return self._cache_size

    @property
    def max_iter(self) -> int:
        return self._max_iter

    # ===========================================================
    # Setters
    @kernel.setter
    def kernel(self, kernel):
        self._kernel = kernel

    @degree.setter
    def degree(self, degree: int):
        assert degree >= 0
        self._degree = degree

    @gamma.setter
    def gamma(self, gamma):
        self._gamma = gamma

    @coef0.setter
    def coef0(self, coef0: float):
        self._coef0 = coef0

    @tol.setter
    def tol(self, tol: float):
        self._tol = tol

    @C.setter
    def C(self, C: float):
        assert C > 0
        self._C = C

    @epsilon.setter
    def epsilon(self, epsilon: float):
        assert epsilon > 0
        self._epsilon = epsilon

    @shrinking.setter
    def shrinking(self, shrinking: bool):
        self._shrinking = shrinking

    @cache_size.setter
    def cache_size(self, cache_size: float):
        self._cache_size = cache_size

    @max_iter.setter
    def max_iter(self, max_iter: int):
        assert max_iter >= -1
        self._max_iter = max_iter



from sklearn.linear_model import LinearRegression as LinearRegressor


class LinearRegression:
    def __init__(self, parameters: dict = None):
        # Model parameters
        self._fit_intercept = True
        self._copy_X = True
        self._n_jobs = None
        self._positive = False

        # Change if user provided changes in dictionary
        if parameters != None:
            for key, value in parameters.items():
                setattr(self, key, value)

    # ===========================================================
    # Methods
    def regressor(self):
        return LinearRegressor(
            fit_intercept=self._fit_intercept,
            copy_X=self._copy_X,
            n_jobs=self._n_jobs,
            positive=self._positive,
        )

    # ===========================================================
    # Getters
    @property
    def fit_intercept(self) -> bool:
        return self._fit_intercept

    @property
    def copy_X(self) -> bool:
        return self._copy_X

    @property
    def n_jobs(self) -> int:
        return self._n_jobs

    @property
    def positive(self) -> bool:
        return self._positive

    # ===========================================================
    # Setters
    @fit_intercept.setter
    def fit_intercept(self, fit_intercept: bool):
        self._fit_intercept = fit_intercept

    @copy_X.setter
    def copy_X(self, copy_X: bool):
        self._copy_X = copy_X

    @n_jobs.setter
    def n_jobs(self, n_jobs: int):
        assert n_jobs >= -1
        self.n_jobs = n_jobs

    @positive.setter
    def positive(self, positive: bool):
        self.positive = positive
